import h5py
import scipy.io as sio
import numpy as np
from torch.utils.data import Dataset
import torch
from sklearn.preprocessing import StandardScaler, MinMaxScaler, MaxAbsScaler


class HL_Data:

    def __init__(self):
        self.f = None
        self.data = None
        self.paddingData = None
        self.dataName = "HL_Data"

    def getFullIndex(self):
        shape = self.data.shape
        x, y = np.meshgrid(np.arange(shape[1]), np.arange(shape[2]))
        x, y = x.flatten(), y.flatten()
        self.full_index = np.stack([x, y]).transpose((1, 0))
        return self.full_index

    def normalize(self, norm):
        if not self.data is None and not norm is None:
            shape = self.data.shape
            self.data = np.reshape(self.data, (shape[0], shape[1] * shape[2]))
            if norm == "stand":
                self.data = StandardScaler().fit_transform(self.data)
            elif norm == "maxmin":
                self.data = MinMaxScaler().fit_transform(self.data)
            elif norm == "maxabs":
                self.data = MaxAbsScaler().fit_transform(self.data)
            self.data = np.reshape(self.data, (shape[0], shape[1], shape[2]))

    def getData(self):
        return self.data

    def getSampleIndex(self, sample_num, iteration, train=True, addData=True):

        # the index from matlab start with 1, not 0, so in the python enviroment, the index need to minus 1
        index = np.array(self.f["train_select_index"])[sample_num - 1, iteration].flatten() - 1
        if train:
            sample = self.f["trainall"][:, index] - 1
        else:
            sample = self.f["trainall"] - 1

        coordinate_index = []
        for i in range(sample.shape[1]):
            x = sample[0, i] // self.data.shape[2]
            y = sample[0, i] % self.data.shape[2]
            coordinate_index.append([x, y, sample[1, i]])

        coordinate_index = np.array(coordinate_index)

        return coordinate_index

    def getPatch(self, x, y, r=2):
        shape = self.data.shape
        if self.paddingData is None:
            self.paddingData = np.zeros((shape[0], shape[1]+2*r, shape[2]+2*r))
            self.paddingData[:, r:-r, r:-r] = self.data
        assert (x>=0 and x<shape[1]), "the second dimension of coordinate is out of index"
        assert (y>=0 and y<shape[2]), "the third dimension of coordinate is out of index"
        x_s, x_e = x, x+r+r+1
        y_s, y_e = y, y+r+r+1
        result = self.paddingData[:, x_s:x_e, y_s:y_e]
        return result

    def getClassNum(self):
        return len(np.unique(self.f["trainall"][1]))


class Indian_Pines_data(HL_Data):
    def __init__(self, norm=None):
        super(Indian_Pines_data, self).__init__()
        self.f = sio.loadmat("/home/kinfai/Code/3DGCNN_34/data/Indian_Pines_data.mat")
        self.data = self.f['data'].transpose(2, 1, 0)
        self.normalize(norm)
        self.dataName = "Indian_Pines_data"


class PaviaU_data(HL_Data):
    def __init__(self, norm=None):
        super(PaviaU_data, self).__init__()
        self.f = sio.loadmat("/home/kinfai/Code/3DGCNN_34/data/PaviaU_data.mat")
        self.data = self.f['data'].transpose(2, 1, 0)
        self.normalize(norm)
        self.dataName = "PaviaU_data"


class XiongAn_data(HL_Data):
    def __init__(self, norm=None):
        super(XiongAn_data, self).__init__()
        self.d = h5py.File("/home/kinfai/Code/3DGCNN_34/data/XiongAn_data.mat")
        self.f = sio.loadmat("/home/kinfai/Code/3DGCNN_34/data/XiongAn_trte_perc.mat")
        self.data = np.array(self.d['data'])
        self.normalize(norm)
        self.dataName = "XiongAn_data_perc"


class Houston2018_data(HL_Data):
    def __init__(self, norm=None):
        super(Houston2018_data, self).__init__()
        self.f = sio.loadmat("/home/kinfai/Code/3DGCNN_34/data/Houston2018_data_perc.mat")
        self.data = self.f['data'].transpose(2, 1, 0)
        self.normalize(norm)
        self.dataName = "Houston2018_data_perc"


class GF5_data(HL_Data):
    def __init__(self, norm=None):
        super(GF5_data, self).__init__()
        self.d = h5py.File("/home/kinfai/Code/3DGCNN_34/data/gf5_data.mat")
        self.f = sio.loadmat("/home/kinfai/Code/3DGCNN_34/data/gf5_trte_perc.mat")
        self.data = np.array(self.d['data'])
        self.normalize(norm)
        self.dataName = "gf5_data_perc"


class HLDataSet(Dataset):

    def __init__(self, data, sam_num=3, ite=0, r=2, train=True, useFullIndex=False):
        assert isinstance(data, HL_Data), "the data is not an instance of HLData."
        self.data = data
        self.train = train
        if useFullIndex:
            self.train = False
            self.sample_index = data.getFullIndex()
        else:
            self.sample_index = data.getSampleIndex(sam_num, ite, train=train)
        self.r = r

        self.useFullIndex = useFullIndex


    def __len__(self):
        return self.sample_index.shape[0]

    def getPatchShape(self):
        return (self.data.data.shape[0], self.r*2+1, self.r*2+1)

    def __getitem__(self, item):
        sam = self.sample_index[item]
        if self.train:
            return torch.as_tensor(self.data.getPatch(sam[0], sam[1], r=self.r), dtype=torch.float).unsqueeze(0), sam[2]
        else:
            if self.useFullIndex:
                return torch.as_tensor(self.data.getPatch(sam[0], sam[1], r=self.r), dtype=torch.float).unsqueeze(0), sam[0], sam[1]
            else:
                return torch.as_tensor(self.data.getPatch(sam[0], sam[1], r=self.r), dtype=torch.float).unsqueeze(0), sam[2], sam[0], sam[1]


# test code
if __name__ == "__main__":

    data = PaviaU_data()
    # print(data.getClassNum())
    dataset = HLDataSet(data, useFullIndex=True)
    print(dataset[10])
    # print(dataset.getPatchShape())
    # for i in dataset:
    #     print(type(i))
    # print(data.getFullIndex())

    # data = Italy_data_hl()
    # for ite in range(10):
    #     for sam_num in range(3, 15):
    #         for sam in data.getSampleIndex(sam_num, ite):
    #             print(data.getPatch(sam[0], sam[1]).shape)
