import numpy as np
import torch
from torch import nn
import torch.nn.functional as F
from torch.autograd import Function
from torch.nn.modules.utils import _pair
from torch.nn.modules.conv import _ConvNd
from torch.autograd.function import once_differentiable
from GCNN3d import _C


class GOF_Function(Function):
    @staticmethod
    def forward(ctx, weight, gaborFilterBank):
        ctx.save_for_backward(weight, gaborFilterBank)
        output = _C.gof_forward(weight, gaborFilterBank)
        return output

    @staticmethod
    @once_differentiable
    def backward(ctx, grad_output):
        weight, gaborFilterBank = ctx.saved_tensors
        grad_weight = _C.gof_backward(grad_output, gaborFilterBank)
        return grad_weight, None


class MConv(_ConvNd):
    '''
    Base layer class for modulated convolution
    '''

    def __init__(self, in_channels, out_channels, kernel_size, M=13, nScale=1, stride=1,
                 padding=0, dilation=1, groups=1, bias=True, expand=False, padding_mode='zeros'):
        if groups != 1:
            raise ValueError('Group-conv not supported!')
        kernel_size = (M,) + _pair(kernel_size)
        stride = _pair(stride)
        padding = _pair(padding)
        dilation = _pair(dilation)
        super(MConv, self).__init__(
            in_channels, out_channels, kernel_size, stride, padding, dilation,
            False, _pair(0), groups, bias, padding_mode)
        self.expand = expand
        self.M = M
        self.need_bias = bias
        self.generate_MFilters(nScale, kernel_size)
        self.GOF_Function = GOF_Function.apply

    def generate_MFilters(self, nScale, kernel_size):
        raise NotImplementedError

    def forward(self, x):
        if self.expand:
            x = self.do_expanding(x)
        new_weight = self.GOF_Function(self.weight, self.MFilters)
        new_bias = self.expand_bias(self.bias) if self.need_bias else self.bias
        # if self.padding_mode == 'circular':
        #     expanded_padding = ((self.padding[1] + 1) // 2, self.padding[1] // 2,
        #                         (self.padding[0] + 1) // 2, self.padding[0] // 2)
        #     return F.conv2d(F.pad(input, expanded_padding, mode='circular'),
        #                     self.weight, self.bias, self.stride,
        #                     _pair(0), self.dilation, self.groups)
        return F.conv2d(x, new_weight, new_bias, self.stride,
                        self.padding, self.dilation, self.groups)

    def do_expanding(self, x):
        index = []
        # for i in range(x.size(1)):
        #     for _ in range(self.M):
        #         index.append(i)
        for i in range(x.size(1) - x.size(1) % self.M):
            index.append(i)
        index = torch.LongTensor(index).cuda() if x.is_cuda else torch.LongTensor(index)
        return x.index_select(1, index)

    def expand_bias(self, bias):
        index = []
        for i in range(bias.size()):
            for _ in range(self.M):
                index.append(i)
        index = torch.LongTensor(index).cuda() if bias.is_cuda else torch.LongTensor(index)
        return bias.index_select(0, index)


class GConv3d(MConv):
    '''
    Gabor Convolutional Operation Layer
    '''

    def __init__(self, in_channels, out_channels, kernel_size, M=13, nScale=1, stride=1,
                 padding=0, dilation=1, groups=1, bias=False, expand=False, padding_mode='zeros'):
        super(GConv3d, self).__init__(in_channels, out_channels, kernel_size, M, nScale, stride,
                                      padding, dilation, groups, bias, expand, padding_mode)

    def generate_MFilters(self, nScale, kernel_size):
        # To generate Gabor Filters
        self.register_buffer('MFilters', get_3d_gabor_filter_bank(nScale))
        # self.register_buffer('MFilters', get_3d_gabor_filter_bank(nScale, *kernel_size))


def get_3d_gabor_filter(f, theta, phi, ratio, x, y, d):
    sigma = ratio / f

    u = f * np.sin(theta) * np.cos(phi)
    v = f * np.sin(theta) * np.sin(phi)
    w = f * np.cos(theta)

    [X, Y, Z] = np.meshgrid(np.arange(-x // 2 + 1, x // 2 + 1),
                            np.arange(-y // 2 + 1, y // 2 + 1),
                            np.arange(-d // 2 + 1, d // 2 + 1))

    prefix = 1 / (((2 * np.pi) ** 1.5) * sigma ** 3)
    gaussian = prefix * np.exp(-(X ** 2 + Y ** 2 + Z ** 2) / (2 * sigma ** 2))
    # modulate = np.exp(1j*2*np.pi*(u*X + v*Y + w*Z))
    cosine = np.cos(2 * np.pi * (u * X + v * Y + w * Z))

    # g = gaussian * modulate
    g_real = gaussian * cosine
    g_real = np.swapaxes(np.swapaxes(g_real, 0, 2), 1, 2)

    return g_real


def get_3d_gabor_filter_bank(nScale=1, M=13, x=3, y=3, d=13):
    f = 1 / (2 * nScale)

    theta = np.array([0, 1 / 4 * np.pi, 1 / 2 * np.pi, 3 / 4 * np.pi])
    phi = np.array([0, 1 / 4 * np.pi, 1 / 2 * np.pi, 3 / 4 * np.pi])

    g_filter_bank = np.zeros([M, d, x, y])
    counter = 0
    for i in range(4):
        for j in range(4):
            g_filter_bank[counter] = get_3d_gabor_filter(f=f, theta=theta[i], phi=phi[j], ratio=1, x=x, y=y, d=d)
            counter += 1
            if i == 0:
                break

    g_filter_bank = torch.from_numpy(np.float32(g_filter_bank))

    return g_filter_bank
